# Dense
from __future__ import annotations

CASEID = "_caseid_"
ALTID = "_altid_"

# Sparse
CASEALT = "_casealt_"
ALTIDX = "_alt_idx_"
CASEPTR = "_caseptr_"
GROUPID = "_groupid_"
INGROUP = "_ingroup_"
